from flask import Blueprint, jsonify  # abort, url_for

import sciolyid.config as config
from sciolyid.data import groups, master_id_list
from sciolyid.web.config import logger

bp = Blueprint("about", __name__, url_prefix="/about")


@bp.route("/list", methods=("GET",))
def list_id_items():
    logger.info("endpoint: about.list")
    if config.options["id_groups"]:
        output = groups
    else:
        output = {"items": master_id_list}
    return jsonify(output)


@bp.route("/info", methods=("GET",))
def info():
    logger.info("endpoint: about.info")
    return jsonify(
        {
            "idName": config.options["id_type"],
            "description": config.options["bot_description"],
            "server": config.options["support_server"],
            "verificationServer": (
                None
                if config.options["verification_server"]
                == config.options["support_server"]
                else config.options["verification_server"]
            ),
            "source": config.options["source_link"],
        }
    )



import hashlib
import json
import os
import shutil
import time
import urllib.parse
from io import BytesIO
from itertools import chain
from typing import Set, Union

import requests
from flask import Blueprint, abort, jsonify, request, send_file
from PIL import Image

import sciolyid.config as config
import sciolyid.web.functions.webhooks as webhooks
import sciolyid.web.tasks.git_tasks as git_tasks
from sciolyid.data import master_id_list
from sciolyid.web.config import logger
from sciolyid.web.functions.images import (find_duplicates, generate_id_lookup,
                                           verify_image)
from sciolyid.web.functions.user import fetch_profile, get_user_id
from sciolyid.web.git import verify_repo
from sciolyid.web.tasks import database

bp = Blueprint("upload", __name__, url_prefix="/upload")


def add_images(
    sources: list,
    destinations: Union[str, list],
    user_id: str,
    username: str,
    use_filenames: bool = True,
):
    different_dests = False
    if isinstance(destinations, list):
        if len(destinations) != len(sources):
            logger.info("source/dest invalid")
            raise IndexError("sources and destinations are not the same length")
        different_dests = True

    verify_repo.remote("origin").pull()
    items: Set[str] = set()
    for i, item in enumerate(sources):
        filename = item.split("/")[-1] if use_filenames else ""
        destination_path = (
            config.options["validation_local_dir"] + destinations[i]
            if different_dests
            else destinations
        )
        os.makedirs(destination_path, exist_ok=True)
        shutil.copyfile(item, destination_path + filename)
        items.add(item.split("/")[-2])

    webhooks.send("add", user_id=user_id, num=len(sources), items=list(items))
    git_tasks.push.delay(f"add images: id-{user_id}\n\nUsername: {username}", user_id)


@bp.route("/", methods=("GET", "POST"))
def upload_files():
    logger.info("endpoint: upload")

    if request.method == "GET":
        return """
            <h1>Upload new File</h1>
            <form method=post enctype=multipart/form-data>
            <input type=file name="file" multiple>
            <input type=hidden name="item" value="dinosaur">
            <input type=submit value=Upload>
            </form>
            """

    user_id: str = get_user_id()
    if not request.files:
        abort(415, "Missing Files")
    if len(request.files) > 10:
        abort(413, "You can only upload 10 files at a time!")
    item = request.form["item"]
    if item not in master_id_list:
        abort(400, "item is invalid")
    files = chain.from_iterable(request.files.listvalues())
    output: dict = {"invalid": [], "duplicates": {}, "sha1": {}, "rejected": []}
    id_lookup = generate_id_lookup()
    for upload in files:
        save_path = f"{config.options['tmp_upload_dir']}{user_id}/{item}/"
        os.makedirs(save_path, exist_ok=True)
        tmp_path = f"{save_path}tmp"
        upload.save(tmp_path)
        with open(tmp_path, "rb") as f:
            sha1 = hashlib.sha1(f.read()).hexdigest()
            if sha1 in id_lookup.values():
                output["rejected"].append(upload.filename)
                os.remove(tmp_path)
                continue

            f.seek(0)
            ext = verify_image(f, upload.mimetype)
            if not ext:
                output["invalid"].append(upload.filename)
                os.remove(tmp_path)
                continue

            dupes = find_duplicates(Image.open(f))
            if dupes:
                output["duplicates"][upload.filename] = dupes

        output["sha1"][upload.filename] = sha1
        os.rename(tmp_path, f"{save_path}{sha1}.{ext}")

    return jsonify(output)


@bp.route("/delete/<string:image_id>", methods=("DELETE",))
def delete(image_id):
    logger.info("endpoint: upload.delete")
    user_id: str = get_user_id()
    tmp = f"{config.options['tmp_upload_dir']}{user_id}/"
    if not os.path.exists(tmp):
        abort(404, "no uploaded images")
    images = []
    for directory in os.listdir(tmp):
        if os.path.isdir(tmp + directory):
            images += list(
                map(lambda x, d=directory: (x, d), os.listdir(tmp + directory))
            )
    found = False
    for filename in images:
        if os.path.splitext(filename[0])[0] == image_id:
            found = os.path.join(filename[1], filename[0])
            break
    if not found:
        abort(404, "image id not found!")
    os.remove(tmp + found)
    if not os.listdir(tmp + os.path.split(found)[0]):
        os.rmdir(tmp + os.path.split(found)[0])
    return jsonify({"deleted": True})


@bp.route("/save", methods=("GET", "POST"))
def save():
    logger.info("endpoint: upload.save")
    user_id: str = get_user_id()

    if database.exists(f"sciolyid.upload.save:{user_id}"):
        abort(400, "save already in progress!")
    database.set(f"sciolyid.upload.save:{user_id}", "1")
    database.delete(f"sciolyid.upload.status:{user_id}")
    database.hset(
        f"sciolyid.upload.status:{user_id}",
        mapping={"start": int(time.time()), "status": json.dumps(["IN_PROGRESS"])},
    )

    username: str = fetch_profile(user_id)["username"]

    save_path: str = config.options["tmp_upload_dir"] + user_id + "/"
    if not os.path.exists(save_path):
        database.delete(
            f"sciolyid.upload.save:{user_id}", f"sciolyid.upload.status:{user_id}"
        )
        abort(404, "No images uploaded")

    sources: list = []
    destinations: list = []
    for directory in os.listdir(save_path):
        remote_path: str = config.options["validation_repo_dir"] + directory + "/"
        current_path: str = save_path + directory + "/"
        for filename in os.listdir(current_path):
            sources.append(current_path + filename)
            destinations.append(remote_path)

    add_images(sources, destinations, user_id, username)
    shutil.rmtree(save_path)

    status: dict = database.hgetall(f"sciolyid.upload.status:{user_id}")
    status = {x[0].decode(): json.loads(x[1].decode()) for x in status.items()}
    return jsonify(status)


@bp.route("/status", methods=("GET",))
def upload_status():
    logger.info("endpoint: upload.status")
    user_id: str = get_user_id()
    if not database.exists(f"sciolyid.upload.status:{user_id}"):
        abort(404, "no current save")
    status: dict = database.hgetall(f"sciolyid.upload.status:{user_id}")
    status = {x[0].decode(): json.loads(x[1].decode()) for x in status.items()}
    return jsonify(status)


@bp.route("/uploaded", methods=("GET",))
def uploaded():
    logger.info("endpoint: upload.uploaded")
    user_id: str = get_user_id()
    save_path: str = config.options["tmp_upload_dir"] + user_id + "/"
    if not os.path.exists(save_path) or not len(os.listdir(save_path)) > 0:
        abort(404, "No uploaded files")
    output: dict = dict()
    for directory in os.listdir(save_path):
        output[directory] = []
        for filename in os.listdir(save_path + directory + "/"):
            output[directory].append(filename)
    return jsonify(output)


@bp.route("/image/<path:image_path>", methods=("GET",))
def send_image(image_path: str):
    logger.info("endpoint: upload.send_image")
    user_id: str = get_user_id()
    save_path: str = os.path.abspath(config.options["tmp_upload_dir"] + user_id)
    item, filename = os.path.split(image_path)
    if not os.path.exists(save_path) or item not in os.listdir(save_path):
        abort(404, "item not found")
    if filename not in os.listdir(os.path.join(save_path, item)):
        abort(404, "filename not found")
    return send_file(os.path.join(save_path, item, filename))


@bp.route("/remote", methods=("GET",))
def get_remote_image():
    logger.info("endpoint: upload.get_remote_image")
    get_user_id()
    url: str = request.args.get("url", "", str)
    if not url:
        abort(400, "No url passed.")
    parsed = urllib.parse.urlparse(url)
    if not parsed.scheme or not parsed.netloc:
        abort(400, "Invalid Url")
    requests.get(urllib.parse.urlunparse(parsed), timeout=10)
    resp = requests.get(url, timeout=10)
    if resp.status_code != 200:
        abort(404, "No image found at that location.")
    content = BytesIO(resp.content)
    ext = verify_image(content, resp.headers.get("content-type"))
    if ext is False:
        abort(404, "Invalid image")
    content.seek(0)
    return send_file(content, attachment_filename=f"file.{ext}")



import os

import celery.bin.beat
import celery.bin.worker
import redis
from celery import Celery

import sciolyid.config as config

database = redis.from_url(os.environ[config.options["celery_broker_env"]])

celery_app = Celery(
    "sciolyid.web.tasks",
    broker=os.environ[config.options["celery_broker_env"]],
    # backend=os.environ[config.options["celery_broker_env"]],
    include=("sciolyid.web.tasks.git_tasks"),
)

CHECK_INTERVAL = 60 * 15
celery_app.conf.beat_schedule = {
    "update_repos": {
        "task": "sciolyid.web.tasks.git_tasks.move_images",
        "schedule": CHECK_INTERVAL,
        "args": tuple(),
    },
}


worker = celery.bin.worker.worker(app=celery_app)
beat = celery.bin.beat.beat(app=celery_app)


def run_worker(args: list):
    worker.run_from_argv("celery", args, "worker")


def run_beat(args: list):
    beat.run_from_argv(
        "celery",
        args
        + [
            f"--schedule={config.options['bot_files_dir']}celerybeat-schedule",
            f"--max-interval={CHECK_INTERVAL + 10}",
        ],
        "beat",
    )


### Database communication definitions ###

# sciolyid.upload.save:{user_id} = 1  # set if save in progress
# sciolyid.upload.status:{user_id} = {
#     start:  # start time
#     end:  # end time, 0 if in progress
#     opcode: op_code  # human readable
#     cur_count: cur_count
#     max_count: max_count
#     message: message
#     status:  # "IN_PROGRESS" during, push result flags after
# }


# sciolyid.verify.images:valid = {  # sorted set
#     image_id: num of times
# }
# sciolyid.verify.images:invalid = {  # sorted set
#     image_id: num of times
# }
# sciolyid.verify.images:duplicate = {  # sorted set
#     image_id: num of times
# }

# sciolyid.verify.user:{user_id} = [  # set
#     image_id...
# ]



import os
import re
import time

import authlib
from authlib.integrations.flask_client import OAuth
from flask import (Blueprint, abort, jsonify, make_response, redirect, request,
                   session, url_for)
from sentry_sdk import capture_exception

import sciolyid.config as config
from sciolyid.web.config import FRONTEND_URL, app, logger
from sciolyid.web.functions.user import fetch_profile, get_user_id, fetch_server_id

bp = Blueprint("user", __name__, url_prefix="/user")
oauth = OAuth(app)

relative_url_regex = re.compile(
    r"/[^/](?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))*"
)

DISCORD_CLIENT_SECRET: str = os.getenv(config.options["client_secret_env"], "")
oauth.register(
    name="discord",
    client_id=config.options["client_id"],
    client_secret=DISCORD_CLIENT_SECRET,
    access_token_url="https://discord.com/api/oauth2/token",
    access_token_params=None,
    authorize_url="https://discord.com/api/oauth2/authorize",
    authorize_params=None,
    api_base_url="https://discord.com/api/",
    client_kwargs={"scope": "identify guilds", "prompt": "consent"},
)
discord = oauth.discord


@bp.route("/login", methods=("GET",))
def login():
    logger.info("endpoint: login")
    redirect_uri = url_for("user.authorize", _external=True, _scheme="https")
    resp = make_response(oauth.discord.authorize_redirect(redirect_uri))
    redirect_after: str = request.args.get("redirect", FRONTEND_URL, str)
    if (
        relative_url_regex.fullmatch(redirect_after) is not None
        and len(redirect_after) <= 50
    ):
        resp.headers.add(
            "Set-Cookie",
            "redirect="
            + redirect_after
            + "; Max-Age=180; SameSite=Lax; HttpOnly; Secure",
        )
    else:
        resp.headers.add(
            "Set-Cookie", "redirect=/; Max-Age=180; SameSite=Lax; HttpOnly; Secure"
        )
    return resp


@bp.route("/authorize")
def authorize():
    logger.info("endpoint: authorize")
    redirect_uri = url_for("user.authorize", _external=True, _scheme="https")
    oauth.discord.authorize_access_token(redirect_uri=redirect_uri)
    user_profile: dict = oauth.discord.get("users/@me").json()
    user_guilds: dict = oauth.discord.get("users/@me/guilds").json()

    if str(fetch_server_id()) not in (guild["id"] for guild in user_guilds):
        return redirect(config.options["verification_server"])

    session["uid"] = user_profile["id"]
    session["date"] = str(int(time.time()))

    redirect_cookie: str = request.cookies.get("redirect")
    if relative_url_regex.fullmatch(redirect_cookie) is not None:
        redirection = FRONTEND_URL + redirect_cookie
    else:
        redirection = FRONTEND_URL + "/"
    return redirect(redirection)


@bp.route("/logout", methods=("GET",))
def logout():
    logger.info("endpoint: logout")

    session.pop("uid", None)
    session.pop("date", None)

    redirect_after: str = request.args.get("redirect", FRONTEND_URL, str)
    if relative_url_regex.fullmatch(redirect_after) is not None:
        redirect_url = FRONTEND_URL + redirect_after
    else:
        redirect_url = FRONTEND_URL
    return redirect(redirect_url)


@bp.route("/profile", methods=("GET",))
def get_profile():
    logger.info("endpoint: profile")
    uid = get_user_id()
    profile = fetch_profile(uid)
    return jsonify(profile)


@app.errorhandler(authlib.common.errors.AuthlibBaseError)
def handle_authlib_error(e):
    logger.info(f"error with oauth login: {e}")
    capture_exception(e)
    return jsonify(error="An error occurred with the login"), 500



# sessions.py | commands for sessions
# Copyright (C) 2019-2020  EraserBird, person_v1.32, hmmm

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import time

import discord
from discord.ext import commands

import sciolyid.config as config
from sciolyid.data import database, groups, logger
from sciolyid.functions import CustomCooldown


class Sessions(commands.Cog):
    def __init__(self, bot):
        self.bot = bot

    def _get_options(self, ctx):
        bw, group, wiki, strict = database.hmget(
            f"session.data:{ctx.author.id}", ["bw", "group", "wiki", "strict"]
        )
        options = (
            f"**Black & White:** {bw==b'bw'}\n"
            + (
                f"**{config.options['category_name']}:** {group.decode('utf-8') if group else 'None'}\n"
                if config.options["id_groups"]
                else ""
            )
            + f"**Wiki Embeds**: {wiki==b'wiki'}\n"
            + f"**Strict Spelling**: {strict==b'strict'}"
        )

        return options

    def _get_stats(self, ctx):
        start, correct, incorrect, total = map(
            int,
            database.hmget(
                f"session.data:{ctx.author.id}", ["start", "correct", "incorrect", "total"],
            ),
        )
        elapsed = str(datetime.timedelta(seconds=round(time.time()) - start))
        try:
            accuracy = round(100 * (correct / (correct + incorrect)), 2)
        except ZeroDivisionError:
            accuracy = 0

        stats = (
            f"**Duration:** `{elapsed}`\n"
            + f"**# Correct:** {correct}\n"
            + f"**# Incorrect:** {incorrect}\n"
            + f"**Total:** {total}\n"
            + f"**Accuracy:** {accuracy}%\n"
        )
        return stats

    async def _send_stats(self, ctx, preamble):
        database_key = f"session.incorrect:{ctx.author.id}"

        embed = discord.Embed(type="rich", colour=discord.Color.blurple(), title=preamble)
        embed.set_author(name=config.options["bot_signature"])

        if database.zcard(database_key) != 0:
            leaderboard_list = database.zrevrangebyscore(
                database_key, "+inf", "-inf", 0, 5, True
            )
            leaderboard = ""

            for i, stats in enumerate(leaderboard_list):
                leaderboard += f"{i+1}. **{stats[0].decode('utf-8')}** - {int(stats[1])}\n"
        else:
            logger.info(f"no items in {database_key}")
            leaderboard = f"**There are no missed {config.options['id_type']}.**"

        embed.add_field(name="Options", value=self._get_options(ctx), inline=False)
        embed.add_field(name="Stats", value=self._get_stats(ctx), inline=False)
        embed.add_field(
            name=f"Top Missed {config.options['id_type'].title()}",
            value=leaderboard,
            inline=False,
        )

        await ctx.send(embed=embed)

    @commands.group(
        brief=f"- Base session command. Use '{config.options['prefixes'][0]}help session' for more info.",
        help="- Base session command\nSessions will record your activity for an amount of time and "
        + "will give you stats on how your performance and "
        + "also set global variables such as black and white"
        + (" or specific categories." if config.options["id_groups"] else "."),
        aliases=["ses", "sesh"],
    )
    async def session(self, ctx):
        if ctx.invoked_subcommand is None:
            await ctx.send(
                "**Invalid subcommand passed.**\n*Valid Subcommands:* `start, view, stop`"
            )

    # starts session
    @session.command(
        brief="- Starts session",
        help="- Starts session.\n"
        + f"Arguments passed will become the default arguments to '{config.options['prefixes'][0]}{config.options['id_type'][:-1]}', "
        + "but can be manually overwritten during use.\n"
        + f"These settings can be changed at any time with '{config.options['prefixes'][0]}session edit', "
        + "and arguments can be passed in any order.\n",
        aliases=["st"],
        usage=("[bw] [category]" if config.options["id_groups"] else "[bw]"),
    )
    @commands.check(CustomCooldown(3.0, bucket=commands.BucketType.user))
    async def start(self, ctx, *, args_str: str = ""):
        logger.info("command: start session")

        if database.exists(f"session.data:{ctx.author.id}"):
            logger.info("already session")
            await ctx.send(
                f"**There is already a session running.** *Change settings/view stats with `{config.options['prefixes'][0]}session edit`*"
            )
            return
        else:
            args = args_str.lower().split(" ")
            logger.info(f"args: {args}")

            if "bw" in args:
                bw = "bw"
            else:
                bw = ""

            if "wiki" in args:
                wiki = ""
            else:
                wiki = "wiki"

            if "strict" in args:
                strict = "strict"
            else:
                strict = ""

            group_args = []
            for category in set(
                list(groups.keys())
                + [
                    item
                    for group in groups.keys()
                    for item in config.options["category_aliases"][group]
                ]
            ).intersection({arg.lower() for arg in args}):
                if category not in groups.keys():
                    category = next(
                        key
                        for key, value in config.options["category_aliases"].items()
                        if category in value
                    )
                group_args.append(category)
            if group_args and config.options["id_groups"]:
                group = " ".join(group_args).strip()
            else:
                group = ""

            logger.info(f"adding bw: {bw}; group: {group}; wiki: {wiki}; strict: {strict}")

            database.hmset(
                f"session.data:{ctx.author.id}",
                {
                    "start": round(time.time()),
                    "stop": 0,
                    "correct": 0,
                    "incorrect": 0,
                    "total": 0,
                    "bw": bw,
                    "group": group,
                    "wiki": wiki,
                    "strict": strict,
                },
            )
            await ctx.send(f"**Session started with options:**\n{self._get_options(ctx)}")

    # views session
    @session.command(
        brief="- Views session",
        help="- Views session\nSessions will record your activity for an amount of time and "
        + "will give you stats on how your performance and "
        + "also set global variables such as black and white"
        + (" or specific categories." if config.options["id_groups"] else "."),
        aliases=["view"],
        usage=("[bw] [category]" if config.options["id_groups"] else "[bw]"),
    )
    @commands.check(CustomCooldown(3.0, bucket=commands.BucketType.user))
    async def edit(self, ctx, *, args_str: str = ""):
        logger.info("command: view session")

        if database.exists(f"session.data:{ctx.author.id}"):
            args = args_str.lower().split(" ")
            logger.info(f"args: {args}")

            if "bw" in args:
                if not database.hget(f"session.data:{ctx.author.id}", "bw"):
                    logger.info("adding bw")
                    database.hset(f"session.data:{ctx.author.id}", "bw", "bw")
                else:
                    logger.info("removing bw")
                    database.hset(f"session.data:{ctx.author.id}", "bw", "")

            if "wiki" in args:
                if database.hget(f"session.data:{ctx.author.id}", "wiki"):
                    logger.info("disabling wiki embeds")
                    database.hset(f"session.data:{ctx.author.id}", "wiki", "")
                else:
                    logger.info("enabling wiki embeds")
                    database.hset(f"session.data:{ctx.author.id}", "wiki", "wiki")

            if "strict" in args:
                if database.hget(f"session.data:{ctx.author.id}", "strict"):
                    logger.info("disabling strict spelling")
                    database.hset(f"session.data:{ctx.author.id}", "strict", "")
                else:
                    logger.info("enabling strict spelling")
                    database.hset(f"session.data:{ctx.author.id}", "strict", "strict")

            group_args = []
            for category in set(
                list(groups.keys())
                + [
                    item
                    for group in groups.keys()
                    for item in config.options["category_aliases"][group]
                ]
            ).intersection({arg.lower() for arg in args}):
                if category not in groups.keys():
                    category = next(
                        key
                        for key, value in config.options["category_aliases"].items()
                        if category in value
                    )
                group_args.append(category)
            if group_args and config.options["id_groups"]:
                toggle_group = list(group_args)
                current_group = (
                    database.hget(f"session.data:{ctx.author.id}", "group")
                    .decode("utf-8")
                    .split(" ")
                )
                add_group = []
                logger.info(f"toggle group: {toggle_group}")
                logger.info(f"current group: {current_group}")
                for o in set(toggle_group).symmetric_difference(set(current_group)):
                    add_group.append(o)
                logger.info(f"adding groups: {add_group}")
                database.hset(
                    f"session.data:{ctx.author.id}", "group", " ".join(add_group).strip(),
                )
            await self._send_stats(ctx, f"**Session started previously.**\n")
        else:
            await ctx.send(
                f"**There is no session running.** *You can start one with `{config.options['prefixes'][0]}session start`*"
            )

    # stops session
    @session.command(help="- Stops session", aliases=["stp", "end"])
    @commands.check(CustomCooldown(3.0, bucket=commands.BucketType.user))
    async def stop(self, ctx):
        logger.info("command: stop session")

        if database.exists(f"session.data:{ctx.author.id}"):
            database.hset(f"session.data:{ctx.author.id}", "stop", round(time.time()))

            await self._send_stats(ctx, "**Session stopped.**\n")
            database.delete(f"session.data:{ctx.author.id}")
            database.delete(f"session.incorrect:{ctx.author.id}")
        else:
            await ctx.send(
                f"**There is no session running.** *You can start one with `{config.options['prefixes'][0]}session start`*"
            )


def setup(bot):
    bot.add_cog(Sessions(bot))



import csv
import imghdr
import os
from typing import Dict, Optional, Set, Union

import imagehash
import requests
from flask import abort
from PIL import Image

import sciolyid.config as config
from sciolyid.web.config import logger

VALID_MIMETYPES = ("image/jpeg", "image/png")
VALID_IMG_TYPES = ("jpeg", "png")
MAX_FILESIZE = 4000000  # 4 mb


def find_duplicates(image, distance: int = 5, ignore_verify=False) -> list:
    logger.info("find duplicates")
    files: Set[str] = set()
    for url in config.options["hashes_url"]:
        if (
            ignore_verify
            and "/".join(config.options["validation_repo_url"].split("/")[-2:-1]).split(
                ".git"
            )[0]
            in url
        ):
            continue
        resp = requests.get(url, timeout=10)
        if resp.status_code != 200:
            logger.info(
                f"hashes lookup failed: status {resp.status_code}; url {resp.url}"
            )
            return ["Failed to get hashes file."]
        files = files.union(set(map(lambda x: x.strip(), resp.text.split("\n"))))
    files.discard("")

    if isinstance(image, str):
        image = Image.open(image)
    current_hash = imagehash.phash(image)
    matches = []
    r = csv.reader(files)
    for url, image_hash in r:
        if current_hash - imagehash.hex_to_hash(image_hash) <= distance:
            matches.append(url)
    return matches


def generate_id_lookup() -> Optional[Dict[str, str]]:
    logger.info("generate id lookup")
    files: Set[str] = set()
    for url in config.options["ids_url"]:
        resp = requests.get(url, timeout=10)
        if resp.status_code != 200:
            logger.info(f"id lookup failed: status {resp.status_code}; url {resp.url}")
            return None
        files = files.union(set(map(lambda x: x.strip(), resp.text.split("\n"))))
    files.discard("")

    lookup = {}
    r = csv.reader(files)
    for filename, image_id in r:
        lookup[filename] = image_id
    return lookup


def filename_lookup(start_path: str) -> dict:
    id_lookup = generate_id_lookup()
    if not id_lookup:
        abort(404, "filename lookup failed!")
    result = {}
    stack = []
    stack.append(start_path)
    while stack:
        current = stack.pop()
        for child_filename in os.listdir(current):
            child_path = current + "/" + child_filename
            if os.path.isdir(child_path):
                stack.append(child_path)
                continue
            if imghdr.what(child_path) in VALID_IMG_TYPES:
                image_id = id_lookup.get("./" + os.path.relpath(child_path, start_path))
                if image_id:
                    result[image_id] = child_path
    return result


def verify_image(f, mimetype) -> Union[bool, str]:
    if mimetype not in VALID_MIMETYPES:
        return False

    f.seek(0, 2)
    size = f.tell()
    f.seek(0)
    if size > MAX_FILESIZE:
        return False

    ext = imghdr.what(None, h=f.read())
    if ext not in VALID_IMG_TYPES:
        return False

    try:
        Image.open(f).verify()
    except:  # pylint: disable=bare-except
        return False

    return ext
